@@ -42,6 +42,7 @@ enum {
 enum {
 	CHANGE_PAGE,
 	FINISHED,
+        SIGNAL_EXTERNAL_LINK,
 	N_SIGNALS
 };
 
@@ -56,6 +57,8 @@ struct _EvViewPresentation
 {
 	GtkWidget base;
 
+        guint                  is_constructing : 1;
+
 	guint                  current_page;
 	cairo_surface_t       *current_surface;
 	EvDocument            *document;
@@ -94,15 +97,15 @@ struct _EvViewPresentationClass
 	GtkWidgetClass base_class;
 
 	/* signals */
-	void (* change_page) (EvViewPresentation *pview,
-			      GtkScrollType       scroll);
-	void (* finished)    (EvViewPresentation *pview);
+	void (* change_page)   (EvViewPresentation *pview,
+                                GtkScrollType       scroll);
+	void (* finished)      (EvViewPresentation *pview);
+        void (* external_link) (EvViewPresentation *pview,
+                                EvLinkAction       *action);
 };
 
 static guint signals[N_SIGNALS] = { 0 };
 
-static void ev_view_presentation_next_page               (EvViewPresentation *pview);
-static void ev_view_presentation_previous_page           (EvViewPresentation *pview);
 static void ev_view_presentation_set_cursor_for_location (EvViewPresentation *pview,
 							  gdouble             x,
 							  gdouble             y);
@@ -111,18 +114,19 @@ static void ev_view_presentation_set_cursor_for_location (EvViewPresentation *pv
 
 G_DEFINE_TYPE (EvViewPresentation, ev_view_presentation, GTK_TYPE_WIDGET)
 
+static GdkRGBA black = { 0., 0., 0., 1. };
+static GdkRGBA white = { 1., 1., 1., 1. };
+
 static void
 ev_view_presentation_set_normal (EvViewPresentation *pview)
 {
 	GtkWidget *widget = GTK_WIDGET (pview);
-	GtkStyle  *style;
 
 	if (pview->state == EV_PRESENTATION_NORMAL)
 		return;
 
 	pview->state = EV_PRESENTATION_NORMAL;
-	style = gtk_widget_get_style (widget);
-	gdk_window_set_background (gtk_widget_get_window (widget), &style->black);
+	gdk_window_set_background_rgba (gtk_widget_get_window (widget), &black);
 	gtk_widget_queue_draw (widget);
 }
 
@@ -130,14 +134,12 @@ static void
 ev_view_presentation_set_black (EvViewPresentation *pview)
 {
 	GtkWidget *widget = GTK_WIDGET (pview);
-	GtkStyle  *style;
 
 	if (pview->state == EV_PRESENTATION_BLACK)
 		return;
 
 	pview->state = EV_PRESENTATION_BLACK;
-	style = gtk_widget_get_style (widget);
-	gdk_window_set_background (gtk_widget_get_window (widget), &style->black);
+	gdk_window_set_background_rgba (gtk_widget_get_window (widget), &black);
 	gtk_widget_queue_draw (widget);
 }
 
@@ -145,14 +147,12 @@ static void
 ev_view_presentation_set_white (EvViewPresentation *pview)
 {
 	GtkWidget *widget = GTK_WIDGET (pview);
-	GtkStyle  *style;
 
 	if (pview->state == EV_PRESENTATION_WHITE)
 		return;
 
 	pview->state = EV_PRESENTATION_WHITE;
-	style = gtk_widget_get_style (widget);
-	gdk_window_set_background (gtk_widget_get_window (widget), &style->white);
+	gdk_window_set_background_rgba (gtk_widget_get_window (widget), &white);
 	gtk_widget_queue_draw (widget);
 }
 
@@ -381,6 +381,25 @@ ev_view_presentation_delete_job (EvViewPresentation *pview,
 	g_object_unref (job);
 }
 
+static void
+ev_view_presentation_reset_jobs (EvViewPresentation *pview)
+{
+        if (pview->curr_job) {
+                ev_view_presentation_delete_job (pview, pview->curr_job);
+                pview->curr_job = NULL;
+        }
+
+        if (pview->prev_job) {
+                ev_view_presentation_delete_job (pview, pview->prev_job);
+                pview->prev_job = NULL;
+        }
+
+        if (pview->next_job) {
+                ev_view_presentation_delete_job (pview, pview->next_job);
+                pview->next_job = NULL;
+        }
+}
+
 static void
 ev_view_presentation_update_current_page (EvViewPresentation *pview,
 					  guint               page)
@@ -485,7 +504,7 @@ ev_view_presentation_update_current_page (EvViewPresentation *pview,
 		gtk_widget_queue_draw (GTK_WIDGET (pview));
 }
 
-static void
+void
 ev_view_presentation_next_page (EvViewPresentation *pview)
 {
 	guint n_pages;
@@ -510,7 +529,7 @@ ev_view_presentation_next_page (EvViewPresentation *pview)
 		ev_view_presentation_update_current_page (pview, new_page);
 }
 
-static void
+void
 ev_view_presentation_previous_page (EvViewPresentation *pview)
 {
 	gint new_page = 0;
@@ -735,6 +754,9 @@ ev_view_presentation_link_is_supported (EvViewPresentation *pview,
 	case EV_LINK_ACTION_TYPE_GOTO_DEST:
 		return ev_link_action_get_dest (action) != NULL;
 	case EV_LINK_ACTION_TYPE_NAMED:
+        case EV_LINK_ACTION_TYPE_GOTO_REMOTE:
+        case EV_LINK_ACTION_TYPE_EXTERNAL_URI:
+        case EV_LINK_ACTION_TYPE_LAUNCH:
 		return TRUE;
 	default:
 		return FALSE;
@@ -793,14 +815,15 @@ ev_view_presentation_get_link_at_location (EvViewPresentation *pview,
 }
 
 static void
-ev_vew_presentation_goto_link_dest (EvViewPresentation *pview,
-				    EvLink             *link)
+ev_vew_presentation_handle_link (EvViewPresentation *pview,
+                                 EvLink             *link)
 {
 	EvLinkAction *action;
 
 	action = ev_link_get_action (link);
 
-	if (ev_link_action_get_action_type (action) == EV_LINK_ACTION_TYPE_NAMED) {
+        switch (ev_link_action_get_action_type (action)) {
+	case EV_LINK_ACTION_TYPE_NAMED: {
 		const gchar *name = ev_link_action_get_name (action);
 
 		if (g_ascii_strcasecmp (name, "FirstPage") == 0) {
@@ -815,13 +838,25 @@ ev_vew_presentation_goto_link_dest (EvViewPresentation *pview,
 			n_pages = ev_document_get_n_pages (pview->document);
 			ev_view_presentation_update_current_page (pview, n_pages - 1);
 		}
-	} else {
+        }
+                break;
+
+	case EV_LINK_ACTION_TYPE_GOTO_DEST: {
 		EvLinkDest *dest;
 		gint        page;
 
 		dest = ev_link_action_get_dest (action);
 		page = ev_document_links_get_dest_page (EV_DOCUMENT_LINKS (pview->document), dest);
 		ev_view_presentation_update_current_page (pview, page);
+        }
+                break;
+        case EV_LINK_ACTION_TYPE_GOTO_REMOTE:
+        case EV_LINK_ACTION_TYPE_EXTERNAL_URI:
+        case EV_LINK_ACTION_TYPE_LAUNCH:
+                g_signal_emit (pview, signals[SIGNAL_EXTERNAL_LINK], 0, action);
+                break;
+        default:
+                break;
 	}
 }
 
@@ -846,7 +881,7 @@ ev_view_presentation_set_cursor (EvViewPresentation *pview,
 	gdk_window_set_cursor (gtk_widget_get_window (widget), cursor);
 	gdk_flush ();
 	if (cursor)
-		gdk_cursor_unref (cursor);
+		g_object_unref (cursor);
 }
 
 static void
@@ -913,21 +948,7 @@ ev_view_presentation_dispose (GObject *object)
 	ev_view_presentation_animation_cancel (pview);
 	ev_view_presentation_transition_stop (pview);
 	ev_view_presentation_hide_cursor_timeout_stop (pview);
-
-	if (pview->curr_job) {
-		ev_view_presentation_delete_job (pview, pview->curr_job);
-		pview->curr_job = NULL;
-	}
-
-	if (pview->prev_job) {
-		ev_view_presentation_delete_job (pview, pview->prev_job);
-		pview->prev_job = NULL;
-	}
-
-	if (pview->next_job) {
-		ev_view_presentation_delete_job (pview, pview->next_job);
-		pview->next_job = NULL;
-	}
+        ev_view_presentation_reset_jobs (pview);
 
 	if (pview->current_surface) {
 		cairo_surface_destroy (pview->current_surface);
@@ -949,11 +970,19 @@ ev_view_presentation_dispose (GObject *object)
 }
 
 static void
-ev_view_presentation_size_request (GtkWidget      *widget,
-				   GtkRequisition *requisition)
+ev_view_presentation_get_preferred_width (GtkWidget *widget,
+                                          gint      *minimum,
+                                          gint      *natural)
 {
-	requisition->width = 0;
-	requisition->height = 0;
+        *minimum = *natural = 0;
+}
+
+static void
+ev_view_presentation_get_preferred_height (GtkWidget *widget,
+                                           gint      *minimum,
+                                           gint      *natural)
+{
+        *minimum = *natural = 0;
 }
 
 static void
@@ -984,15 +1013,8 @@ ev_view_presentation_draw_end_page (EvViewPresentation *pview,
 	area.width = allocation.width;
 	area.height = allocation.height;
 
-        gtk_paint_layout (gtk_widget_get_style (widget),
-                          cr,
-                          gtk_widget_get_state (widget),
-                          FALSE,
-                          widget,
-                          NULL,
-                          15,
-                          15,
-                          layout);
+        gtk_render_layout (gtk_widget_get_style_context (widget),
+                           cr, 15, 15, layout);
 
 	pango_font_description_free (font_desc);
 	g_object_unref (layout);
@@ -1156,7 +1178,7 @@ ev_view_presentation_button_release_event (GtkWidget      *widget,
 								  event->x,
 								  event->y);
 		if (link)
-			ev_vew_presentation_goto_link_dest (pview, link);
+			ev_vew_presentation_handle_link (pview, link);
 		else
 			ev_view_presentation_next_page (pview);
 	}
@@ -1217,10 +1239,9 @@ init_presentation (GtkWidget *widget)
 static void
 ev_view_presentation_realize (GtkWidget *widget)
 {
-	GdkWindow    *window;
-	GtkStyle     *style;
-	GdkWindowAttr attributes;
-	GtkAllocation allocation;
+	GdkWindow     *window;
+	GdkWindowAttr  attributes;
+	GtkAllocation  allocation;
 
 	gtk_widget_set_realized (widget, TRUE);
 
@@ -1250,10 +1271,8 @@ ev_view_presentation_realize (GtkWidget *widget)
 
 	gdk_window_set_user_data (window, widget);
 	gtk_widget_set_window (widget, window);
-
-	gtk_widget_style_attach (widget);
-	style = gtk_widget_get_style (widget);
-	gdk_window_set_background (window, &style->black);
+        gtk_style_context_set_background (gtk_widget_get_style_context (widget),
+                                          window);
 
 	g_idle_add ((GSourceFunc)init_presentation, widget);
 }
@@ -1333,7 +1352,7 @@ ev_view_presentation_set_property (GObject      *object,
 		pview->current_page = g_value_get_uint (value);
 		break;
 	case PROP_ROTATION:
-		pview->rotation = g_value_get_uint (value);
+                ev_view_presentation_set_rotation (pview, g_value_get_uint (value));
 		break;
 	case PROP_INVERTED_COLORS:
 		pview->inverted_colors = g_value_get_boolean (value);
@@ -1343,6 +1362,23 @@ ev_view_presentation_set_property (GObject      *object,
 	}
 }
 
+static void
+ev_view_presentation_get_property (GObject    *object,
+                                   guint       prop_id,
+                                   GValue     *value,
+                                   GParamSpec *pspec)
+{
+        EvViewPresentation *pview = EV_VIEW_PRESENTATION (object);
+
+        switch (prop_id) {
+        case PROP_ROTATION:
+                g_value_set_uint (value, ev_view_presentation_get_rotation (pview));
+                break;
+        default:
+                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+        }
+}
+
 static GObject *
 ev_view_presentation_constructor (GType                  type,
 				  guint                  n_construct_properties,
@@ -1355,6 +1391,7 @@ ev_view_presentation_constructor (GType                  type,
 										  n_construct_properties,
 										  construct_params);
 	pview = EV_VIEW_PRESENTATION (object);
+        pview->is_constructing = FALSE;
 
 	if (EV_IS_DOCUMENT_LINKS (pview->document)) {
 		pview->page_cache = ev_page_cache_new (pview->document);
@@ -1370,12 +1407,14 @@ ev_view_presentation_class_init (EvViewPresentationClass *klass)
 	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 	GObjectClass   *gobject_class = G_OBJECT_CLASS (klass);
 	GtkBindingSet  *binding_set;
+        GtkCssProvider *provider;
 
 	klass->change_page = ev_view_presentation_change_page;
 
         gobject_class->dispose = ev_view_presentation_dispose;
 
-	widget_class->size_request = ev_view_presentation_size_request;
+	widget_class->get_preferred_width = ev_view_presentation_get_preferred_width;
+	widget_class->get_preferred_height = ev_view_presentation_get_preferred_height;
 	widget_class->realize = ev_view_presentation_realize;
         widget_class->draw = ev_view_presentation_draw;
 	widget_class->key_press_event = ev_view_presentation_key_press_event;
@@ -1386,6 +1425,7 @@ ev_view_presentation_class_init (EvViewPresentationClass *klass)
 
 	gobject_class->constructor = ev_view_presentation_constructor;
 	gobject_class->set_property = ev_view_presentation_set_property;
+        gobject_class->get_property = ev_view_presentation_get_property;
 
 	g_object_class_install_property (gobject_class,
 					 PROP_DOCUMENT,
@@ -1409,8 +1449,8 @@ ev_view_presentation_class_init (EvViewPresentationClass *klass)
 							    "Rotation",
 							    "Current rotation angle",
 							    0, 360, 0,
-							    G_PARAM_WRITABLE |
-							    G_PARAM_CONSTRUCT_ONLY));
+							    G_PARAM_READWRITE |
+							    G_PARAM_CONSTRUCT));
 	g_object_class_install_property (gobject_class,
 					 PROP_INVERTED_COLORS,
 					 g_param_spec_boolean ("inverted_colors",
@@ -1438,6 +1478,15 @@ ev_view_presentation_class_init (EvViewPresentationClass *klass)
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE, 0,
 			      G_TYPE_NONE);
+        signals[SIGNAL_EXTERNAL_LINK] =
+                g_signal_new ("external-link",
+                              G_TYPE_FROM_CLASS (gobject_class),
+                              G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+                              G_STRUCT_OFFSET (EvViewPresentationClass, external_link),
+                              NULL, NULL,
+                              g_cclosure_marshal_VOID__OBJECT,
+                              G_TYPE_NONE, 1,
+                              G_TYPE_OBJECT);
 
 	binding_set = gtk_binding_set_by_class (klass);
 	add_change_page_binding_keypad (binding_set, GDK_KEY_Left,  0, GTK_SCROLL_PAGE_BACKWARD);
@@ -1468,12 +1517,23 @@ ev_view_presentation_class_init (EvViewPresentationClass *klass)
 	gtk_binding_entry_add_signal (binding_set, GDK_KEY_K, 0,
 				      "change_page", 1,
 				      GTK_TYPE_SCROLL_TYPE, GTK_SCROLL_PAGE_BACKWARD);
+
+        provider = gtk_css_provider_new ();
+        gtk_css_provider_load_from_data (provider,
+                                         "EvViewPresentation {\n"
+                                         " background-color: black; }",
+                                         -1, NULL);
+        gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
+                                                   GTK_STYLE_PROVIDER (provider),
+                                                   GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
+        g_object_unref (provider);
 }
 
 static void
 ev_view_presentation_init (EvViewPresentation *pview)
 {
 	gtk_widget_set_can_focus (GTK_WIDGET (pview), TRUE);
+        pview->is_constructing = TRUE;
 }
 
 GtkWidget *
@@ -1498,3 +1558,31 @@ ev_view_presentation_get_current_page (EvViewPresentation *pview)
 {
 	return pview->current_page;
 }
+
+void
+ev_view_presentation_set_rotation (EvViewPresentation *pview,
+                                   gint                rotation)
+{
+        if (rotation >= 360)
+                rotation -= 360;
+        else if (rotation < 0)
+                rotation += 360;
+
+        if (pview->rotation == rotation)
+                return;
+
+        pview->rotation = rotation;
+        g_object_notify (G_OBJECT (pview), "rotation");
+        if (pview->is_constructing)
+                return;
+
+        pview->scale = 0;
+        ev_view_presentation_reset_jobs (pview);
+        ev_view_presentation_update_current_page (pview, pview->current_page);
+}
+
+guint
+ev_view_presentation_get_rotation (EvViewPresentation *pview)
+{
+        return pview->rotation;
+}
